/******************************************************************** 
********************************************************************* 
**                                                                 **
**   this file contains the type definition of the internal token  **
**   representation structure                                      **
**                                                                 **
********************************************************************* 
********************************************************************/ 

 typedef int bool;

 typedef struct tokn
 {
   int          token_type;                /* token type                 */
   int          point_size;                /* point size                 */
   int          font      ;                /* font number                */
   char         font_name[3];              /* font name                  */
   bool         lr        ;                /* font direction             */
   int          width     ;                /* character width            */
   int          vertical_pos;              /* vertical position          */
   int          horizontal_pos;            /* horizontal position        */
   bool         begining  ;                /* begining of word indicator */
   bool         ending    ;                /* end of word indicator      */
   int          fillers_num;               /* how many fillers to put    */
   int          filler_width;              /* the filler width           */
   int		letter_stretch;		   /* internal stretch of letter */
   char         char1     ;                /* 1st byte of char token     */
   char         char2     ;                /* 2nd byte of char token     */
   char         char3     ;                /* 3rd byte of char token     */
   struct tokn  *next     ;                /* next token                 */

   bool         arabic_persian_letter;     /* if it's one of those       */
   char         internal_stretchability;   /* N or S                     */
   char         connectivity;              /* U, N, P, A, or B           */
   int		already_stretched_amount;  /* non-zero if character is   */
					   /* manually stretched         */
   int          next_space_width;          /* if ending is TRUE, then    */
                                           /* is non zero and is the     */
                                           /* width of the space after   */
                                           /* the word                   */
   int          prev_space_width;          /* if begining is TRUE, then  */
                                           /* is non zero and is the     */
                                           /* width of the space before  */
                                           /* the word                   */
   int          next_space_spread;         /* if ending is TRUE, then    */
                                           /* is non zero and is the     */
                                           /* spread of the space after  */
                                           /* the word                   */
   int          prev_space_spread;         /* if begining is TRUE, then  */
                                           /* is non zero and is the     */
                                           /* spread of the space before */
                                           /* the word                   */
 } TOKENTYPE, *TOKENPTR;

 typedef struct gap
 {
   TOKENPTR     before_token;              /* point to token before gap  */
   TOKENPTR     after_token;               /* point to token after  gap  */
   int          before_font;               /* font of token before gap   */
   int          after_font;                /* font of token after gap    */
   int          before_point_size;         /* size of token before gap   */
   int          after_point_size;          /* size of token after gap    */
   int          before_trial_space_width;  /* space width according to   */
                                           /* token before gap           */
   int          after_trial_space_width;   /* space width according to   */
                                           /* token after gap            */
   int          before_char_width;         /* width of token before gap  */
   int          actual_movement;           /* movement over gap          */
   bool         both_trial_space_widths_the_same;
   bool         definite;                  /* gap space width determined */
   int          final_space_width;         /* gap space width            */
   int          final_space_spread;        /* spread added to gap        */

 } GAPTYPE, *GAPPTR;


/*************************************************************************/
/*************************************************************************/

